import requests
from rich.console import Console
console = Console()


class Mixin:
    def ql_get_rewards_for_account(self, ql_account_id):
        query = "query {"
        query += f'rewardMetricsForAccount (period:LAST30_DAYS, accountId: "{ql_account_id}")'
        query += """
                    {
                    buckets {
                        y_SumRewards
                        x_Time
                        bucketWidth
                    }
                    sumRewardAmount
                    }
                }
        """
        try:
            r = requests.post(self.graphql_url, json={'query': query})
            if r.status_code == 200:
                return r.json()['data']['rewardMetricsForAccount']
            else:
                console.log(r.text)
        except Exception as e:
            console.log(query, e)
            return None

    def ql_request_account_rewards(self, ql_id: str):
        query = "query {"
        # query += f'LAST_HOUR:    rewardMetricsForAccount(accountId: "{ql_id}" period: LAST_HOUR) {{ sumRewardAmount }}'
        query += f'LAST24_HOURS: rewardMetricsForAccount(accountId: "{ql_id}" period: LAST24_HOURS) {{ sumRewardAmount }}'
        query += f'LAST7_DAYS:   rewardMetricsForAccount(accountId: "{ql_id}" period: LAST7_DAYS) {{ sumRewardAmount }}'
        query += f'LAST30_DAYS:  rewardMetricsForAccount(accountId: "{ql_id}" period: LAST30_DAYS) {{ sumRewardAmount }}'
        # query += f'LAST_YEAR:    rewardMetricsForAccount(accountId: "{ql_id}" period: LAST_YEAR) {{ sumRewardAmount }}'
        
        query += '}' # query
        try:
            r = requests.post(self.graphql_url, json={'query': query})
            if r.status_code == 200:
                account_rewards = {
                    # 'LAST_HOUR':    r.json()['data']['LAST_HOUR'],
                    'LAST24_HOURS': r.json()['data']['LAST24_HOURS'],
                    'LAST7_DAYS':   r.json()['data']['LAST7_DAYS'],
                    'LAST30_DAYS':  r.json()['data']['LAST30_DAYS'],
                    # 'LAST_YEAR':    r.json()['data']['LAST_YEAR'],

                 }
                return account_rewards
            else:
                return None
       
        except Exception as e:
            console.log(query, e)
            return None

            


import requests
from rich.console import Console
console = Console()


class Mixin:
    def ql_request_passive_delegators_for_lookup(self, before: str = None, after: str = None):
        
        query = "query { passiveDelegation  { delegatorCount "
        
        if before == 'first':
            accounts_str  =  f'delegators (first: {self.nodes_request_limit}) '
        
        elif before != 'null':
            accounts_str  =  f'delegators (last: {self.nodes_request_limit}, before: "{before}") '
        elif 'last' in after:
            accounts_str  =  f'delegators ({after}) ' 
        
        elif after != 'null':
            accounts_str  =  f'delegators (first: {self.nodes_request_limit}, after: "{after}") ' 
        else:
            accounts_str = f'delegators (first: {self.nodes_request_limit})'
        query += accounts_str + '{'
        query += self.pageInfo()
        query += """
                    nodes {
                        stakedAmount
                        restakeEarnings
                        accountAddress {
                            asString
                        }
                    }
        """
        query += ' } } }'

        try:
            r = requests.post(self.graphql_url, json={'query': query})
            if r.status_code == 200:
                return r.json()['data']['passiveDelegation']['delegators']['nodes'], \
                    r.json()['data']['passiveDelegation']['delegators']['pageInfo'], \
                    r.json()['data']['passiveDelegation']['delegatorCount'],                           
            else:
                console.log(r.text)
    
        except Exception as e:
            console.log(query, e)
            return [], [], 0

    def ql_get_passive_pool_info(self):
        query = "query { passiveDelegation {"
        query += """
                    apy_30: apy(period: LAST30_DAYS)
                    apy_7: apy(period: LAST7_DAYS)
                    delegatorCount
                    delegatedStake
                    delegatedStakePercentage
                    commissionRates {
                        finalizationCommission
                        transactionCommission
                        bakingCommission
                    }
                } } """
        try:
            r = requests.post(self.graphql_url, json={'query': query})
            if r.status_code == 200:
                return r.json()['data']['passiveDelegation']                    
            else:
                console.log(r.text)
        
        except Exception as e:
            console.log(query, e)
            return None

    def ql_passive_delegation_main(self):
        query = """
                query {
        passiveDelegation {
            delegatedStakePercentage
            delegatedStake
            delegatorCount
            commissionRates {
            finalizationCommission
            bakingCommission
            transactionCommission
            }
        }
        }
        
        """
        try:
            r = requests.post(self.graphql_url, json={'query': query})
            if r.status_code == 200:
                return r.json()['data']['passiveDelegation']
       
        except Exception as e:
            console.log(query, e)


# ruff: noqa: F403, F405, E402
from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *
from ccdexplorer_fundamentals.enums import NET
from enum import Enum
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
import os
import sys

sys.path.append(os.path.dirname("ccdexplorer_fundamentals"))
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient


class Mixin(_SharedConverters):
    def convertTokenomicsV0(self, message) -> CCD_TokenomicsInfo_V0:
        if self.valueIsEmpty(message):
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)

                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)

            return CCD_TokenomicsInfo_V0(**result)

    def convertTokenomicsV1(self, message) -> CCD_TokenomicsInfo_V1:
        if self.valueIsEmpty(message):
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)

                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)

                elif type(value) is MintRate:
                    result[key] = CCD_MintRate(
                        **{"mantissa": value.mantissa, "exponent": value.exponent}
                    )

            return CCD_TokenomicsInfo_V1(**result)

    def get_tokenomics_info(
        self: GRPCClient,
        block_hash: str,
        net: Enum = NET.MAINNET,
    ) -> CCD_TokenomicsInfo:
        prefix = ""
        result = {}
        blockHashInput = self.generate_block_hash_input_from(block_hash)

        grpc_return_value: TokenomicsInfo = self.stub_on_net(
            net, "GetTokenomicsInfo", blockHashInput
        )

        for descriptor in grpc_return_value.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(
                descriptor, grpc_return_value
            )

            if type(value) is TokenomicsInfo.V0:
                result[f"{prefix}{key}"] = self.convertTokenomicsV0(value)
            elif type(value) is TokenomicsInfo.V1:
                result[f"{prefix}{key}"] = self.convertTokenomicsV1(value)

        return CCD_TokenomicsInfo(**result)



from enum import Enum
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *  # noqa: F403


class CredentialElement(Enum):
    firstName = "First Name"
    lastName = "Last Name"
    sex = "Sex"
    dob = "Date of Birth"
    countryOfResidence = "Country of Residence"
    nationality = "Nationatility"
    idDocType = "Identity Document Type"
    idDocNo = "Identity Document Number"
    idDocIssuer = "Identity Document Issuer"
    idDocIssuedAt = "ID Valid from"
    idDocExpiresAt = "ID Valid to"
    nationalIdNo = "National ID number"
    taxIdNo = "Tax ID number"
    lei = "Legal Entity ID"
    legalName = "Legal Name"
    legalJurisdictionCountry = "Legal Jurisdiction Country"
    businessNumber = "Business Number"
    registrationAuthority = "Registration Authority"


class CredentialDocType(Enum):
    na = "0"
    Passport = "1"
    National_ID_Card = "2"
    Driving_License = "3"
    Immigration_Card = "4"


class Credentials:
    """
    This class processes credential information as retrieved from the node.
    """

    def __init__(self):
        pass

    def determine_id_providers(self, ac):
        """
        Input to this method is the output from the node.
        """
        credentials = []
        for key_policy, v in ac.items():
            v: CCD_AccountCredential  # noqa: F405
            if v.initial:
                v = v.initial
                normal = False
            elif v.normal:
                v = v.normal
                normal = True

            c = {
                "ip_identity": v.ip_id,
                "created_at": v.policy.created_at,
                "valid_to": v.policy.valid_to,
            }
            if normal:
                c.update({"cred_id": v.cred_id})

            if len(v.policy.attributes.keys()) > 0:
                policy_attributes = []
                for key_policy, commitmentAttribute in v.policy.attributes.items():
                    value = commitmentAttribute
                    policy_attributes.append(
                        {"key": CredentialElement[key_policy].value, "value": value}
                    )
                c.update({"policy_attributes": policy_attributes})
            if normal:
                if len(v.commitments.attributes.keys()) > 0:
                    commitment_attributes = []
                    for (
                        key_commitment,
                        commitmentAttribute,
                    ) in v.commitments.attributes.items():
                        value = commitmentAttribute
                        commitment_attributes.append(
                            CredentialElement[key_commitment].value
                        )
                    c.update({"commitment_attributes": commitment_attributes})
            credentials.append(c)
        return credentials


class Identity:
    def __init__(self, account_info: CCD_AccountInfo):  # noqa: F405
        if account_info:
            self.credentials = Credentials().determine_id_providers(
                account_info.credentials
            )
            self.threshold = account_info.threshold
        else:
            self.credentials = []
            self.threshold = 0



from pydantic import BaseModel

class ConcordiumReward(BaseModel):
    account: str = None
    poolOwner: int = None
    bakerReward: int
    finalizationReward: int
    transactionFees: int
    tag: str


from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *

# ruff: noqa: F403, F405, E402
from ccdexplorer_fundamentals.enums import NET
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
from typing import Iterator
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *


class Mixin(_SharedConverters):
    def get_instance_list(
        self: GRPCClient,
        block_hash: str,
        net: Enum = NET.MAINNET,
    ) -> list[ContractAddress]:
        result = []
        blockHashInput = self.generate_block_hash_input_from(block_hash)

        grpc_return_value: Iterator[ContractAddress] = self.stub_on_net(
            net, "GetInstanceList", blockHashInput
        )
        for instance in list(grpc_return_value):
            result.append(self.convertType(instance))

        return result



import datetime as dt

class Mixin:
    def get_rewards_data_for_accounts(self, date: str):
        pipeline = [
                {
                    "$match":     {
                        "date": {"$eq": date},
                        "reward.tag": {"$eq": "PaydayAccountReward"}
                     }
                    # '$match': { "account_id": account_id }
                }, 
                # { '$sort': {'blockSlotTime': -1} },
                # {'$limit': limit},
                {
                '$addFields': {
                    'sum_rewards': {
                        '$sum': [
                            '$reward.bakerReward', '$reward.finalizationReward', '$reward.transactionFees'
                        ]
                    }
                }
                },   
                # {'$project': {'_id': 0, 'stakedAmount': 1, 'sum_rewards': 1}}

        ]

        return pipeline
    def get_rewards_data_for_account(self, account_id: str):
        pipeline = [
                {
                    # "$match":     {"timestamp": {"$gt": start_date} },
                    '$match': { "account_id": account_id }
                }, 
                # { '$sort': {'blockSlotTime': -1} },
                # {'$limit': limit},
                {
                '$addFields': {
                    'sum_rewards': {
                        '$sum': [
                            '$reward.bakerReward', '$reward.finalizationReward', '$reward.transactionFees'
                        ]
                    }
                }
                },   
                # {'$project': {'_id': 0, 'stakedAmount': 1, 'sum_rewards': 1}}

        ]

        return pipeline

    def get_rewards_data_for_pool(self, pool_id: str):
        try:
            pool_id = int(pool_id)
        except:
            pool_id = 'passiveDelegation'

        pipeline = [
                # {
                #     "$match":     {"date": {"$in": days_in_apy_period} },
                # },
                {
                    '$match': { "poolOwner": pool_id }
                }, 
                { '$sort': {'blockSlotTime': -1} },
                # {'$limit': limit},
                {
                '$addFields': {
                    'baker_ratio': {
                        '$divide': [
                            '$pool_status.currentPaydayStatus.bakerEquityCapital', 
                            '$pool_status.currentPaydayStatus.effectiveStake'

                        ]
                    }
                }
                },   
                {
                '$addFields': {
                    'delegators_ratio': {
                        '$divide': [
                            '$pool_status.currentPaydayStatus.delegatedCapital', 
                            '$pool_status.currentPaydayStatus.effectiveStake'
                            ]
                    }
                }
                },
                {
                '$addFields': {
                    'delegators_baking_reward': {
                        '$multiply': [
                            {
                                '$subtract': [
                                    1, 
                                    '$pool_status.poolInfo.commissionRates.bakingCommission'
                                ]
                            }
                            ,
                            {
                                '$multiply': 
                                [
                                    '$delegators_ratio', 
                                    '$reward.bakerReward'
                                ]
                            }
                        ]
                    }
                }
                },
                {
                '$addFields': {
                    'delegators_transaction_reward': {
                        '$multiply': [
                            {
                                '$subtract': [
                                    1, 
                                    '$pool_status.poolInfo.commissionRates.transactionCommission'
                                ]
                            }
                            ,
                            {
                                '$multiply': 
                                [
                                    '$delegators_ratio', 
                                    '$reward.transactionFees'
                                ]
                            }
                        ]
                    }
                }
                },   
                {
                '$addFields': {
                    'delegators_finalization_reward': {
                        '$multiply': [
                            {
                                '$subtract': [
                                    1, 
                                    '$pool_status.poolInfo.commissionRates.finalizationCommission'
                                ]
                            }
                            ,
                            {
                                '$multiply': 
                                [
                                    '$delegators_ratio', 
                                    '$reward.finalizationReward'
                                ]
                            }
                        ]
                    }
                }
                },
                {
                '$addFields': {
                    'baker_baking_reward': {
                        '$subtract': [
                            '$reward.bakerReward', 
                            '$delegators_baking_reward'
                        ]
                    }
                }
                },
                {
                '$addFields': {
                    'baker_transaction_reward': {
                        '$subtract': [
                            '$reward.transactionFees', 
                            '$delegators_transaction_reward'
                                ]
                            }
                    }
                },   
                {
                '$addFields': {
                    'baker_finalization_reward': {
                        '$subtract': [
                            '$reward.finalizationReward', 
                            '$delegators_finalization_reward'
                                ]
                            }
                    }
                },   
                {
                '$addFields': {
                    'delegator_reward': {
                        '$add': [
                            '$delegators_baking_reward', 
                            '$delegators_transaction_reward',
                            '$delegators_finalization_reward'
                        ]
                    }
                }
                },   
                {
                '$addFields': {
                    'total_reward': {
                        '$sum': [
                            '$reward.bakerReward', '$reward.finalizationReward', '$reward.transactionFees'
                        ]
                    }
                }
                },   
                {
                '$addFields': {
                    'baker_reward': {
                        '$subtract': [
                            '$total_reward', 
                            '$delegator_reward'
                        ]
                    }
                }
                },   
                # {'$project': {'_id': 0, 'bakerEquityCapital': 1, 'delegatedCapital': 1, 'bakerReward_baker': 1}}

        ]

        return pipeline

    def get_rewards_data_for_pools(self, date: str):
        pipeline = [
                {
                    "$match":     {
                        "date": {"$eq": date},
                        "reward.tag": {"$eq": "PaydayPoolReward"}
                     }
                }, 
                {
                '$addFields': {
                    'baker_ratio': {
                        '$divide': [
                            '$pool_status.currentPaydayStatus.bakerEquityCapital', 
                            '$pool_status.currentPaydayStatus.effectiveStake'

                        ]
                    }
                }
                },   
                {
                '$addFields': {
                    'delegators_ratio': {
                        '$divide': [
                            '$pool_status.currentPaydayStatus.delegatedCapital', 
                            '$pool_status.currentPaydayStatus.effectiveStake'
                            ]
                    }
                }
                },
                {
                '$addFields': {
                    'delegators_baking_reward': {
                        '$multiply': [
                            {
                                '$subtract': [
                                    1, 
                                    '$pool_status.poolInfo.commissionRates.bakingCommission'
                                ]
                            }
                            ,
                            {
                                '$multiply': 
                                [
                                    '$delegators_ratio', 
                                    '$reward.bakerReward'
                                ]
                            }
                        ]
                    }
                }
                },
                {
                '$addFields': {
                    'delegators_transaction_reward': {
                        '$multiply': [
                            {
                                '$subtract': [
                                    1, 
                                    '$pool_status.poolInfo.commissionRates.transactionCommission'
                                ]
                            }
                            ,
                            {
                                '$multiply': 
                                [
                                    '$delegators_ratio', 
                                    '$reward.transactionFees'
                                ]
                            }
                        ]
                    }
                }
                },   
                {
                '$addFields': {
                    'delegators_finalization_reward': {
                        '$multiply': [
                            {
                                '$subtract': [
                                    1, 
                                    '$pool_status.poolInfo.commissionRates.finalizationCommission'
                                ]
                            }
                            ,
                            {
                                '$multiply': 
                                [
                                    '$delegators_ratio', 
                                    '$reward.finalizationReward'
                                ]
                            }
                        ]
                    }
                }
                },
                {
                '$addFields': {
                    'baker_baking_reward': {
                        '$subtract': [
                            '$reward.bakerReward', 
                            '$delegators_baking_reward'
                        ]
                    }
                }
                },
                {
                '$addFields': {
                    'baker_transaction_reward': {
                        '$subtract': [
                            '$reward.transactionFees', 
                            '$delegators_transaction_reward'
                                ]
                            }
                    }
                },   
                {
                '$addFields': {
                    'baker_finalization_reward': {
                        '$subtract': [
                            '$reward.finalizationReward', 
                            '$delegators_finalization_reward'
                                ]
                            }
                    }
                },   
                {
                '$addFields': {
                    'delegator_reward': {
                        '$add': [
                            '$delegators_baking_reward', 
                            '$delegators_transaction_reward',
                            '$delegators_finalization_reward'
                        ]
                    }
                }
                },   
                {
                '$addFields': {
                    'total_reward': {
                        '$sum': [
                            '$reward.bakerReward', '$reward.finalizationReward', '$reward.transactionFees'
                        ]
                    }
                }
                },   
                {
                '$addFields': {
                    'baker_reward': {
                        '$subtract': [
                            '$total_reward', 
                            '$delegator_reward'
                        ]
                    }
                }
                }
        ]

        return pipeline



import requests
from rich.console import Console
console = Console()


class Mixin:
    def ql_request_delegators_for_lookup(self, baker_id: int, before: str = None, after: str = None):
            
            query = "query {"
            query += f'bakerByBakerId (bakerId: {baker_id}) '
            
            query += '{ state { ... on ActiveBakerState { pool { delegatorCount '
            
            if before == 'first':
                accounts_str  =  f'delegators (first: {self.nodes_request_limit}) '
            
            elif before != 'null':
                accounts_str  =  f'delegators (last: {self.nodes_request_limit}, before: "{before}") '
            elif 'last' in after:
                accounts_str  =  f'delegators ({after}) ' 
            
            elif after != 'null':
                accounts_str  =  f'delegators (first: {self.nodes_request_limit}, after: "{after}") ' 
            else:
                accounts_str = f'delegators (first: {self.nodes_request_limit})'
            query += accounts_str + '{'
            query += self.pageInfo()
            query += """
                        nodes {
                            stakedAmount
                            restakeEarnings
                            accountAddress {
                                asString
                            }
                        }
            """
            query += ' } } } } } }'

            try:
                r = requests.post(self.graphql_url, json={'query': query})
                if r.status_code == 200:
                    if r.json()['data']['bakerByBakerId']['state'] != {}:
                        return r.json()['data']['bakerByBakerId']['state']['pool']['delegators']['nodes'], \
                            r.json()['data']['bakerByBakerId']['state']['pool']['delegators']['pageInfo'], \
                            r.json()['data']['bakerByBakerId']['state']['pool']['delegatorCount'],                           
                    else:
                        return [], [], 0
                else:
                    console.log(r.text)
        
            except Exception as e:
                console.log(query, e)
                return None

    def ql_get_baker_pool_info(self, baker_id:int):
        
        query = "query {"
        query += f'bakerByBakerId (bakerId: {baker_id}) '
        query += """
             { state {
            ... on ActiveBakerState {
                pool {
                    apy_30: apy (period:LAST30_DAYS) {
                        bakerApy
                        delegatorsApy
                        totalApy
                    }
                    apy_7: apy (period:LAST7_DAYS) {
                        bakerApy
                        delegatorsApy
                        totalApy
                    }
                    openStatus
                    metadataUrl
                    delegatorCount
                    delegatedStake
                    delegatedStakeCap
                    rankingByTotalStake {
                        rank
                        total
                    }
                    totalStake
                    totalStakePercentage
                    commissionRates {
                        finalizationCommission
                        transactionCommission
                        bakingCommission
                    }
                
                
                }
            }
            }
        }
        }
        
        """
        try:
            r = requests.post(self.graphql_url, json={'query': query})
            if r.status_code == 200:
                if r.json()['data']['bakerByBakerId']['state'] != {}:
                    return r.json()['data']['bakerByBakerId']['state']['pool']                    
                else:
                    return {}
            else:
                console.log(r.text)
        
        except Exception as e:
            console.log(query, e)
            return None

    def ql_get_info_by_bakerId(self, value):
        query = "query {"
        query += f'bakerByBakerId (bakerId: {value})'
        query += """
                    {
                        id
                        state {
                            ... on ActiveBakerState {
                                nodeStatus {
                                    nodeName
                                }
                                pool {
                                    __typename
                                }
                            }
                        }
                        account {
                            address {
                                asString
                            }
                        }
                    }
                }
        """
        try:
            # r = self.timed_request(query)
            r = requests.post(self.graphql_url, json={'query': query})
            if r.status_code == 200:
                return r.json()['data']['bakerByBakerId']
            else:
                return None
        except Exception as e:
            console.log(query, e)
            return None


# ruff: noqa: F403, F405, E402
from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *
from ccdexplorer_fundamentals.enums import NET
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
from typing import TYPE_CHECKING
import grpc
from datetime import timezone

if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *


class Mixin(_SharedConverters):
    def get_baker_earliest_win_time(
        self: GRPCClient,
        baker_id: int,
        net: Enum = NET.MAINNET,
    ) -> CCD_TimeStamp:
        try:
            grpc_return_value = self.stub_on_net(
                net, "GetBakerEarliestWinTime", BakerId(value=baker_id)
            )
        except grpc.RpcError as rpc_error:
            if rpc_error.code() == grpc.StatusCode.UNIMPLEMENTED:
                print("Not implemented on mainnet")
            return None

        win_time = None

        for descriptor in grpc_return_value.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(
                descriptor, grpc_return_value
            )

            if key == "value":
                win_time = dt.datetime.fromtimestamp(value / 1_000).astimezone(
                    tz=timezone.utc
                )

        return win_time
